from tclint.syntax_tree import ArgExpansion, QuotedWord, BracedWord


class CommandArgError(Exception):
    pass


def arg_count(args, parser):
    # TODO: graceful handling of argsub going into things with recursive parsing.
    # if the argsub happens to be "concrete", we can technically do the right
    # thing (although this should probably be flagged as a readability issue...)
    # otherwise, we should flag that the non-concrete argsub is not okay for
    # these cases. however, I think its not okay-ness doesn't need to be absolute, e.g.
    # I think we could allow:
    #
    #  catch {puts "my script"} {*}$catchopts
    #

    arg_count = 0
    for arg in args:
        if isinstance(arg, ArgExpansion):
            if arg.contents is None:
                return None
            arg_count += len(parser.parse_list(arg.contents))
        else:
            arg_count += 1

    return arg_count


def check_count(command, min=None, max=None):
    def check(args, parser):
        if min is None and max is None:
            return None

        count = arg_count(args, parser)
        if count is None:
            return None

        if min == max and count != min:
            raise CommandArgError(
                f"wrong # of args to {command}: got {count}, expected {min}"
            )

        if min is not None and count < min:
            raise CommandArgError(
                f"not enough args to {command}: got {count}, expected at least {min}"
            )

        if max is not None and count > max:
            raise CommandArgError(
                f"too many args to {command}: got {count}, expected no more than {max}"
            )

        return None

    return check


def subcommands(name, subcommands, default=None):
    def check(args, parser):
        try:
            arg = args[0].contents
        except IndexError:
            arg = None

        if arg in subcommands:
            func = subcommands[arg]
            if func is None:
                return None
            new_args = func(args[1:], parser)
            if new_args is None:
                return None
            return args[0:1] + new_args
        elif default is not None:
            return default(args, parser)
        else:
            if arg is not None:
                msg = f"invalid subcommand for {name}: got {arg}"
            else:
                msg = f"no subcommand provided for {name}"

            raise CommandArgError(
                f"{msg}, expected one of {', '.join(subcommands.keys())}"
            )

    return check


def eval(args, parser, command):
    if len(args) > 1 and any(isinstance(arg, (QuotedWord, BracedWord)) for arg in args):
        # Slightly odd restriction, but our syntax tree doesn't have a great way
        # to handle this case. We require each command argument to correspond to
        # one child node, but multiple quoted or braced word arguments can be
        # combined into a single subcommand when interpreted eval-style. This
        # requirement exists to facilitate style checking, if we had a separate
        # CST for style checks and AST for logical checks we may be able to
        # handle it.

        raise CommandArgError(
            f"unable to parse multiple {command} arguments when one includes a braced"
            " or quoted word"
        )

    # Construct the body of the eval taking whitespace into account to ensure we get
    # style checking.

    eval_script = ""
    prev_arg_end_pos = None
    for arg in args:
        contents = arg.contents
        if contents is None:
            # TODO: flag sort of eval-specific violation? Common patterns will
            # often trigger this, and it seems useful to be able to turn it off
            raise CommandArgError(
                f"{command} received an argument with a substitution, unable to parse"
                " its arguments"
            )

        if prev_arg_end_pos is not None:
            if prev_arg_end_pos[0] != arg.line:
                # If we have multiple args on the same line, we know there must be a
                # backslash newline. Add it so the parsing works.
                eval_script += "\\\n" * (arg.line - prev_arg_end_pos[0])
                eval_script += " " * (arg.col - 1)
            else:
                eval_script += " " * (arg.col - prev_arg_end_pos[1])
        eval_script += contents

        prev_arg_end_pos = arg.end_pos

    script = parser.parse(eval_script, pos=(args[0].pos))
    script.end_pos = args[-1].end_pos

    return [script]



import string
import re

from tclint.lexer import (
    Lexer,
    TclSyntaxError,
    TOK_BACKSLASH_NEWLINE,
    TOK_NEWLINE,
    TOK_SEMI,
    TOK_WS,
    TOK_QUOTE,
    TOK_ARG_EXPANSION,
    TOK_LBRACE,
    TOK_RBRACE,
    TOK_LBRACKET,
    TOK_RBRACKET,
    TOK_DOLLAR,
    TOK_LPAREN,
    TOK_RPAREN,
    TOK_HASH,
    TOK_ALPHA_CHARS,
    TOK_NUM_CHARS,
    TOK_NAMESPACE_SEP,
    TOK_EOF,
)
from tclint.syntax_tree import (
    Script,
    Comment,
    Command,
    CommandSub,
    ArgExpansion,
    VarSub,
    BareWord,
    BracedWord,
    QuotedWord,
    CompoundBareWord,
    List,
    Expression,
    BracedExpression,
    ParenExpression,
    UnaryOp,
    BinaryOp,
    TernaryOp,
    Function,
)
from tclint.commands import CommandArgError, get_commands
from tclint.violations import Rule, Violation


def _strip_ws(parse_func):
    """Decorator used by expression parser for stripping whitespace around a node."""

    def func(parser, ts):
        while ts.type() in {TOK_WS, TOK_BACKSLASH_NEWLINE, TOK_NEWLINE}:
            ts.next()

        node = parse_func(parser, ts)

        while ts.type() in {TOK_WS, TOK_BACKSLASH_NEWLINE, TOK_NEWLINE}:
            ts.next()

        return node

    return func


class _Word:
    """Helper class for constructing Word nodes out of multiple segments."""

    def __init__(self):
        self.segments = []
        self.current_segment = ""
        self.current_start = None

    def add_tok(self, tok):
        if self.current_start is None:
            self.current_start = tok.value[1]
        self.current_segment += tok.value[0]

    def add_node(self, node):
        if self.current_segment != "":
            self.segments.append(
                BareWord(self.current_segment, pos=self.current_start, end_pos=node.pos)
            )
            self.current_segment = ""
            self.current_start = None
        self.segments.append(node)

    def resolve(self, end_pos):
        if self.current_segment:
            self.segments.append(
                BareWord(self.current_segment, pos=self.current_start, end_pos=end_pos)
            )

        return self.segments


class Parser:
    def __init__(self, debug=False, command_plugins=None):
        self._debug = debug
        self._debug_indent = 0
        # self._cmd_sub = True indicates that the script will be terminated by ]"""
        self._cmd_sub = False
        # TODO: better way to handle this?
        self.violations = []

        if command_plugins is None:
            command_plugins = []
        self._commands = get_commands(command_plugins)

    def debug(self, *msg):
        if self._debug:
            print("  " * self._debug_indent, end="")
            print(*msg)

    def parse(self, script, pos=None):
        lexer = Lexer(pos=pos)
        lexer.input(script)
        tree = self._parse_script(lexer)

        return tree

    def _parse_command_args(self, routine, args):
        """Since many built-in Tcl commands take in Tcl scripts or expressions
        as arguments, building a complete parse tree requires checking command
        names and possibly parsing their arguments.

        The node of any argument that gets parsed by this method is replaced with
        the parse tree of that argument. Since this process requires checking
        the arguments provided to these commands, this method may report lint
        violations.

        This parsing process is analogous to how the Tcl interpreter interprets
        scripts, and better handles weird edge cases compared to a traditional
        parsing technique. For example, this may look like valid Tcl:

        proc foo {a} {
            # output }
            puts "}"
        }

        But really, it is invalid since the } in the comment terminates the body
        of the proc - Tcl blindly constructs the body of the proc until it
        reaches the first }. tclint handles this correctly.
        """
        if routine not in self._commands:
            return args

        func = self._commands[routine]
        if func is None:
            return args

        try:
            new_args = func(args, self)
        except CommandArgError as e:
            raise e
        except Exception:
            if self.debug:
                raise
            raise CommandArgError(
                f"error parsing command arguments, possibly malformed {routine} command"
            )

        if new_args is None:
            return args

        return new_args

    def parse_script(self, node):
        if node.contents is None:
            raise CommandArgError(
                "expected braced word or word without substitutions in argument"
                " interpreted as script"
            )

        cmd_sub = self._cmd_sub
        self._cmd_sub = False
        script = self.parse(node.contents, pos=node.contents_pos)
        if isinstance(node, BracedWord):
            script.braced = True
        self._cmd_sub = cmd_sub

        script.line = node.line
        script.col = node.col
        script.end_pos = node.end_pos

        return script

    def _parse_script(self, ts):
        self.debug(f"parse_script({ts.current})")
        self._debug_indent += 1
        pos = ts.pos()

        if self._cmd_sub:
            script = CommandSub(pos=pos)
        else:
            script = Script(pos=pos)

        while ts.type() is not TOK_EOF:
            if ts.type() in {TOK_WS, TOK_BACKSLASH_NEWLINE}:
                # strip whitespace at start of command
                ts.next()
                continue

            if ts.type() == TOK_HASH:
                script.add(self.parse_comment(ts))
            else:
                cmd = self.parse_command(ts)
                if cmd is not None:
                    script.add(cmd)

            # when in command sub mode, a script is terminated by ]
            if self._cmd_sub and ts.type() == TOK_RBRACKET:
                return script

            ts.expect(
                TOK_EOF,
                TOK_NEWLINE,
                TOK_SEMI,
                message=f"expected newline or semicolon, got {ts.value()}",
                pos=ts.pos(),
            )

        if self._cmd_sub and ts.type() is TOK_EOF:
            raise TclSyntaxError(
                "reached EOF without finding end of command substitution", pos
            )

        self._debug_indent -= 1

        script.end_pos = ts.pos()

        return script

    def parse_comment(self, ts):
        self.debug(f"parse_comment({ts.current})")
        pos = ts.pos()

        ts.assert_(TOK_HASH)

        value = ""
        while ts.type() not in {TOK_NEWLINE, TOK_EOF}:
            value += ts.value()
            ts.next()

        return Comment(value, pos=pos, end_pos=ts.pos())

    def parse_command(self, ts):
        self.debug(f"parse_command({ts.current})")
        self._debug_indent += 1
        pos = ts.pos()

        routine = self.parse_word(ts)
        if routine is None:
            self._debug_indent -= 1
            return None

        args = []
        while True:
            if ts.type() not in {TOK_WS, TOK_BACKSLASH_NEWLINE}:
                break

            while ts.type() in {TOK_WS, TOK_BACKSLASH_NEWLINE}:
                ts.next()

            word = self.parse_word(ts)
            if word is None:
                break

            args.append(word)

        self._debug_indent -= 1

        try:
            parsed_args = self._parse_command_args(routine.contents, args)
        except CommandArgError as e:
            self.violations.append(Violation(Rule.COMMAND_ARGS, str(e), pos))
            parsed_args = args

        children = [routine, *parsed_args]
        # We need to inherit end pos of last child to prevent us from counting
        # extra whitespace at end of command, which is important for
        # spaces-in-braces check.
        return Command(*children, pos=pos, end_pos=children[-1].end_pos)

    def parse_word(self, ts):
        self.debug(f"parse_word({ts.current})")
        if ts.type() == TOK_ARG_EXPANSION:
            return self.parse_arg_expansion(ts)
        elif ts.type() == TOK_LBRACE:
            return self.parse_braced_word(ts)
        elif ts.type() == TOK_QUOTE:
            return self.parse_quoted_word(ts)
        else:
            return self.parse_bare_word(ts)

    def parse_arg_expansion(self, ts):
        self.debug(f"parse_arg_expansion({ts.current})")
        pos = ts.pos()

        ts.assert_(TOK_ARG_EXPANSION)

        # Arg expansion is just a regular braced word if followed by whitespace
        if ts.type() in {TOK_WS, TOK_BACKSLASH_NEWLINE, TOK_NEWLINE, TOK_EOF}:
            return BracedWord("*", pos=pos, end_pos=ts.pos())

        return ArgExpansion(self.parse_word(ts), pos=pos, end_pos=ts.pos())

    def parse_quoted_word(self, ts):
        self.debug(f"parse_quoted_word({ts.current})")
        self._debug_indent += 1
        pos = ts.pos()

        ts.assert_(TOK_QUOTE)

        word = _Word()
        while ts.type() not in {TOK_QUOTE, TOK_EOF}:
            if ts.type() == TOK_DOLLAR:
                dollar_tok = ts.current
                var_sub = self.parse_var_sub(ts)
                if var_sub:
                    word.add_node(var_sub)
                else:
                    word.add_tok(dollar_tok)
            elif ts.type() == TOK_LBRACKET:
                command_sub = self.parse_command_sub(ts)
                word.add_node(command_sub)
            else:
                word.add_tok(ts.current)
                ts.next()

        res = word.resolve(ts.pos())

        ts.expect(
            TOK_QUOTE, message="reached EOF without finding match for quote", pos=pos
        )

        self._debug_indent -= 1

        if not res:
            res = []

        return QuotedWord(*res, pos=pos, end_pos=ts.pos())

    def parse_braced_word(self, ts):
        self.debug(f"parse_braced_word({ts.current})")
        pos = ts.pos()

        ts.assert_(TOK_LBRACE)

        word = ""
        # store position for each brace we want to match, facilitating good
        # error messages
        expected_braces = [pos]
        while True:
            if ts.type() == TOK_EOF:
                raise TclSyntaxError(
                    "reached EOF without finding match for brace", expected_braces[-1]
                )

            if ts.type() == TOK_LBRACE:
                expected_braces.append(ts.pos())
            elif ts.type() == TOK_RBRACE:
                try:
                    expected_braces.pop()
                except IndexError:
                    raise TclSyntaxError(
                        "found closing brace without matching open brace", ts.pos()
                    )

                if len(expected_braces) == 0:
                    ts.next()
                    break
            word += ts.value()
            ts.next()

        end_pos = ts.pos()
        return BracedWord(word, pos=pos, end_pos=end_pos)

    def parse_bare_word(self, ts):
        self.debug(f"parse_bare_word({ts.current})")
        self._debug_indent += 1
        pos = ts.pos()

        word = _Word()
        delimiters = [TOK_WS, TOK_BACKSLASH_NEWLINE, TOK_NEWLINE, TOK_SEMI, TOK_EOF]

        # In command sub mode, words are ended by ]
        # TODO: consider pushing this down via argument to remove it as a member
        if self._cmd_sub:
            delimiters.append(TOK_RBRACKET)

        while ts.type() not in delimiters:
            if ts.type() == TOK_DOLLAR:
                dollar_tok = ts.current
                var_sub = self.parse_var_sub(ts)
                if var_sub:
                    word.add_node(var_sub)
                else:
                    word.add_tok(dollar_tok)
            elif ts.type() == TOK_LBRACKET:
                command_sub = self.parse_command_sub(ts)
                word.add_node(command_sub)
            else:
                word.add_tok(ts.current)
                ts.next()

        res = word.resolve(ts.pos())

        self._debug_indent -= 1

        if not res:
            return None
        if len(res) == 1:
            return res[0]
        return CompoundBareWord(*res, pos=pos, end_pos=ts.pos())

    def parse_var_sub(self, ts):
        self.debug(f"parse_var_sub({ts.current})")
        pos = ts.pos()

        ts.assert_(TOK_DOLLAR)

        var = ""
        if ts.type() == TOK_LBRACE:
            brace_pos = ts.pos()
            ts.next()
            while ts.type() != TOK_RBRACE:
                if ts.type() is TOK_EOF:
                    raise TclSyntaxError(
                        "reached EOF without finding match for brace", brace_pos
                    )
                var += ts.value()
                ts.next()
            ts.next()

            return VarSub(var, pos=pos, end_pos=ts.pos())

        while ts.type() in {TOK_ALPHA_CHARS, TOK_NUM_CHARS, TOK_NAMESPACE_SEP}:
            var += ts.value()
            ts.next()

        if not var:
            return None

        index_nodes = []
        if ts.type() == TOK_LPAREN:
            paren_pos = ts.pos()
            index = _Word()
            ts.next()
            while ts.type() != TOK_RPAREN:
                if ts.type() == TOK_EOF:
                    raise TclSyntaxError(
                        "reached EOF without finding match for paren", paren_pos
                    )
                if ts.type() == TOK_DOLLAR:
                    dollar_tok = ts.current
                    var_sub = self.parse_var_sub(ts)
                    if var_sub:
                        index.add_node(var_sub)
                    else:
                        index.add_tok(dollar_tok)
                elif ts.type() == TOK_LBRACKET:
                    command_sub = self.parse_command_sub(ts)
                    index.add_node(command_sub)
                else:
                    index.add_tok(ts.current)
                    ts.next()

            index_nodes = index.resolve(ts.pos())
            ts.next()

        var_sub = VarSub(var, pos=pos, end_pos=ts.pos())

        for index_segment in index_nodes:
            var_sub.add(index_segment)

        return var_sub

    def parse_command_sub(self, ts):
        self.debug(f"parse_command_sub({ts.current})")
        self._debug_indent += 1

        pos = ts.pos()
        ts.assert_(TOK_LBRACKET)

        saved_cmd_sub = self._cmd_sub
        self._cmd_sub = True
        script = self._parse_script(ts)
        self._cmd_sub = saved_cmd_sub

        ts.assert_(TOK_RBRACKET)
        end_pos = ts.pos()

        script.line = pos[0]
        script.col = pos[1]
        script.end_pos = end_pos

        self._debug_indent -= 1
        return script

    def parse_list(self, node):
        """Parse contents of node as Tcl list. This is a distinct entry point
        that doesn't get used when generating the main syntax tree, but is used
        in command-specific argument parsing.
        """
        if node.contents is None:
            return None

        # Hack: we know contents start 1 col over, since first char denotes the list
        ts = Lexer(pos=(node.line, node.col + 1))
        ts.input(node.contents)

        DELIMITERS = {TOK_WS, TOK_BACKSLASH_NEWLINE, TOK_NEWLINE}

        list_node = List(pos=node.pos, end_pos=node.end_pos)
        while ts.type() is not TOK_EOF:
            while ts.type() in DELIMITERS:
                ts.next()

            if ts.type() is TOK_EOF:
                break

            if ts.type() == TOK_LBRACE:
                # we can reuse parse_braced_word, since it doesn't use
                # substitutions in any case
                list_node.add(self.parse_braced_word(ts))
            elif ts.type() == TOK_QUOTE:
                quote_word_pos = ts.pos()

                ts.assert_(TOK_QUOTE)

                bare_word_pos = ts.pos()
                contents = ""
                while ts.type() not in {TOK_QUOTE, TOK_EOF}:
                    contents += ts.value()
                    ts.next()
                word = BareWord(contents, pos=bare_word_pos, end_pos=ts.pos())

                ts.expect(
                    TOK_QUOTE,
                    message="reached EOF without finding match for quote",
                    pos=quote_word_pos,
                )

                list_node.add(QuotedWord(word, pos=quote_word_pos, end_pos=ts.pos()))
            else:
                pos = ts.pos()
                contents = ""
                while ts.type() not in {*DELIMITERS, TOK_EOF}:
                    contents += ts.value()
                    ts.next()
                list_node.add(BareWord(contents, pos=pos, end_pos=ts.pos()))

        return list_node

    def parse_expression(self, node):
        if node.contents is None:
            raise CommandArgError(
                "expected braced word or word without substitutions in argument"
                " interpreted as expr"
            )

        ts = Lexer(pos=node.contents_pos)
        ts.input(node.contents)

        contents = self._parse_expression(ts)
        if isinstance(node, BracedWord):
            return BracedExpression(contents, pos=node.pos, end_pos=node.end_pos)

        return Expression(contents, pos=node.pos, end_pos=node.end_pos)

    @_strip_ws
    def _parse_expression(self, ts):
        op1 = self._parse_operand(ts)
        expr = None

        # last condition is hack to break out of expression in case we're in ternary op
        if ts.type() not in {TOK_EOF, TOK_RPAREN} and ts.value() not in {":", ","}:
            if ts.value() == "?":
                expr = TernaryOp(pos=op1.pos)
                expr.add(op1)

                # weird hack to record operator
                n = BareWord("?", pos=ts.pos())
                ts.next()
                n.end_pos = ts.pos()
                expr.add(n)

                op2 = self._parse_expression(ts)
                expr.add(op2)
                if ts.value() != ":":
                    raise TclSyntaxError(
                        "expected ':' to continue ternary expression", ts.pos()
                    )

                # weird hack again
                n = BareWord(":", pos=ts.pos())
                ts.next()
                n.end_pos = ts.pos()
                expr.add(n)

                op3 = self._parse_expression(ts)
                expr.add(op3)
            else:
                expr = BinaryOp(pos=op1.pos)
                expr.add(op1)

                operator = self._parse_operator(ts)
                expr.add(operator)

                op2 = self._parse_expression(ts)
                expr.add(op2)

            if ts.type() != TOK_RPAREN and ts.value() not in {":", ","}:
                ts.expect(TOK_EOF, message="expected end of expression", pos=ts.pos())

        if expr is None:
            return op1

        expr.end_pos = expr.children[-1].end_pos
        return expr

    @_strip_ws
    def _parse_operand(self, ts):
        if ts.type() == TOK_DOLLAR:
            return self.parse_var_sub(ts)
        if ts.type() == TOK_QUOTE:
            return self.parse_quoted_word(ts)
        if ts.type() == TOK_LBRACE:
            return self.parse_braced_word(ts)
        if ts.type() == TOK_LBRACKET:
            return self.parse_command_sub(ts)
        if ts.type() == TOK_LPAREN:
            expr = ParenExpression(pos=ts.pos())
            ts.next()
            expr.add(self._parse_expression(ts))
            ts.expect(
                TOK_RPAREN,
                message="reached EOF without finding match for paren",
                pos=expr.pos,
            )
            expr.end_pos = ts.pos()
            return expr
        if ts.value() in {"-", "+", "~", "!"}:
            operator = ts.value()
            operator_pos = ts.pos()
            ts.next()
            op = UnaryOp(pos=operator_pos)
            op.add(BareWord(operator, pos=operator_pos, end_pos=ts.pos())),
            op.add(self._parse_operand(ts))
            # Since _parse_operand() munches whitespace after the operand, we
            # set the end of the UnaryOp to the end of the operand rather than
            # ts.pos(). Otherwise, the bounds of the UnaryOp would include all
            # that whitespace.
            op.end_pos = op.children[-1].end_pos
            return op

        # If none of these, collect tokens that may comprise an operand
        operand = ""
        operand_pos = ts.pos()

        # First, we want to check for numeric operands (either ints or numeric
        # floats) by consuming tokens as long as they comprise the prefix of a
        # numeric operand
        while ts.type() != TOK_EOF and (
            _is_int_prefix(operand + ts.value())
            or _is_float_prefix(operand + ts.value())
        ):
            operand += ts.value()
            ts.next()

        # Next, we check if we've consumed an entire numeric literal. If so, we
        # move on. If not, we keep consuming tokens that may correspond to a
        # valid bareword (pretty much just alphanumeric chars).
        if not (_is_int_literal(operand) or _is_float_literal(operand)):
            while ts.type() in {TOK_ALPHA_CHARS, TOK_NUM_CHARS}:
                operand += ts.value()
                ts.next()

        # The above method is a little hacky. Note that it doesn't parse things
        # exactly the same as Tcl. E.g. if a script includes `expr {1foo}`,
        # tclint will report an invalid operator "foo", whereas tclsh will
        # report an invalid bareword "1foo". Despite reporting them differently
        # both tools should still catch the same syntax errors, since there are
        # no legal barewords that begin with a numeric literal prefix, and tclsh
        # will stop parsing numeric operands if they're actually followed by a
        # legal operator (e.g. `expr {1eq1}` will be handled properly).

        is_func = _is_function(operand)

        if not (
            _is_int_literal(operand)
            or _is_float_literal(operand)
            or _is_bool_literal(operand)
            or is_func
        ):
            raise TclSyntaxError(
                f"invalid bareword in expression: {operand}", operand_pos
            )

        node = BareWord(operand, pos=operand_pos, end_pos=ts.pos())

        if is_func:
            node = self._parse_function(ts, node)

        return node

    def _parse_operator(self, ts):
        pos = ts.pos()

        # hacky logic to handle parsing legal operators

        if ts.value() in {"*", "&", "|"}:
            # one or two of these characters are legal operators
            operator = ts.value()
            ts.next()
            if ts.value() == operator:
                operator += ts.value()
                ts.next()
        elif ts.value() in {"<", ">"}:
            operator = ts.value()
            ts.next()
            if ts.value() in {operator, "="}:
                operator += ts.value()
                ts.next()
        elif ts.value() in {"=", "!"}:
            operator = ts.value()
            ts.next()
            if ts.value() != "=":
                raise TclSyntaxError(f"invalid operator in expression: {operator}", pos)
            operator += ts.value()
            ts.next()
        elif ts.value() in {"*", "/", "%", "+", "-", "^", "eq", "ne", "in", "ni"}:
            operator = ts.value()
            ts.next()
        else:
            raise TclSyntaxError(f"invalid operator in expression: {ts.value()}", pos)

        return BareWord(operator, pos=pos, end_pos=ts.pos())

    def _parse_function(self, ts, name_node):
        func = Function(pos=name_node.pos)

        func.add(name_node)

        while ts.type() in {TOK_WS, TOK_BACKSLASH_NEWLINE}:
            ts.next()

        ts.expect(
            TOK_LPAREN,
            message="expected open paren after function name",
            pos=name_node.pos,
        )

        delims = {TOK_RPAREN, TOK_EOF}

        if ts.type() not in delims:
            arg = self._parse_expression(ts)
            func.add(arg)

        while ts.type() not in delims:
            if ts.value() != ",":
                raise TclSyntaxError(
                    "expected comma between function arguments", ts.pos()
                )

            # adding comma may seem a little weird since it's non-functional,
            # but this lets us store comma position for style checks
            comma = BareWord(",", pos=ts.pos())
            ts.next()
            comma.end_pos = ts.pos()
            func.add(comma)

            arg = self._parse_expression(ts)
            func.add(arg)

        ts.expect(
            TOK_RPAREN,
            message="expected close paren after function arguments",
            pos=name_node.pos,
        )
        func.end_pos = ts.pos()

        return func


def _all(_list, non_empty=False):
    """Like all(), but if non_empty is True, list must also have at least 1 element."""
    if non_empty and len(_list) == 0:
        return False
    return all(_list)


def _is_int(operand, full=False):
    """Returns whether operand is a valid Tcl integer literal.

    If full is False, will also return True if operand is the prefix of an
    integer literal.  An empty string is not a valid full literal, but is a
    valid prefix.
    """
    # prefixes
    if operand.startswith("0b"):
        return _all([digit in "01" for digit in operand[2:]], non_empty=full)
    if operand.startswith("0o"):
        return _all(
            [digit in string.octdigits for digit in operand[2:]], non_empty=full
        )
    if operand.startswith("0x"):
        return _all(
            [digit in string.hexdigits for digit in operand[2:]], non_empty=full
        )
    if operand.startswith("0"):
        # fun fact: apparently a lone 0 prefix is interpreted as octal
        return _all(
            [digit in string.octdigits for digit in operand[1:]], non_empty=full
        )

    return _all([digit in string.digits for digit in operand], non_empty=full)


def _is_int_literal(operand):
    return _is_int(operand, full=True)


def _is_int_prefix(operand):
    return _is_int(operand, full=False)


def _is_float_literal(operand):
    if operand.lower() in {"nan", "inf"}:
        return True

    return (
        operand != "" and re.fullmatch(r"\d*\.?\d*([Ee][+-]?\d+)?", operand) is not None
    )


def _is_float_prefix(operand):
    """Returns whether operand is the prefix of a valid numeric float literal."""
    return re.fullmatch(r"\d*\.?\d*([Ee][+-]?)?\d*", operand) is not None


def _is_bool_literal(operand):
    return operand in {"false", "no", "off", "true", "yes", "on"}


# map of function names to # arguments accepted
# None indicates 1 or more arguments
# TODO: use these values in an actual separate check. they might want to live elsewhere
_FUNCTIONS = {
    "abs": 1,
    "acos": 1,
    "asin": 1,
    "atan": 1,
    "atan2": 2,
    "bool": 1,
    "ceil": 1,
    "cos": 1,
    "cosh": 1,
    "double": 1,
    "entier": 1,
    "exp": 1,
    "floor": 1,
    "fmod": 2,
    "hyopt": 2,
    "int": 1,
    "isqrt": 1,
    "log": 1,
    "log10": 1,
    "max": None,
    "min": None,
    "pow": 2,
    "rand": 0,
    "round": 1,
    "sin": 1,
    "sinh": 1,
    "sqrt": 1,
    "srand": 1,
    "tan": 1,
    "tanh": 1,
    "wide": 1,
}


def _is_function(operand):
    return operand in _FUNCTIONS.keys()



from enum import Enum
from typing import Tuple


class Rule(Enum):
    """This enum serves a few purposes:

    1) define symbols for rule IDs to be used in code
    2) map these symbols to names in the UI
    3) collect all rule IDs/provide validation for IDs
    """

    INDENT = "indent"
    SPACING = "spacing"
    LINE_LENGTH = "line-length"
    TRAILING_WHITESPACE = "trailing-whitespace"
    COMMAND_ARGS = "command-args"
    REDEFINED_BUILTIN = "redefined-builtin"
    BLANK_LINES = "blank-lines"
    BACKSLASH_SPACING = "backslash-spacing"
    EXPR_FORMAT = "expr-format"
    SPACES_IN_BRACES = "spaces-in-braces"
    UNBRACED_EXPR = "unbraced-expr"

    def __str__(self):
        return self.value


ALL_RULES = [rule for rule in Rule]


class Category(Enum):
    FUNC = "func"
    STYLE = "style"

    def __str__(self):
        return self.value


_CATEGORY_MAP = {
    Rule.INDENT: Category.STYLE,
    Rule.SPACING: Category.STYLE,
    Rule.LINE_LENGTH: Category.STYLE,
    Rule.TRAILING_WHITESPACE: Category.STYLE,
    Rule.BLANK_LINES: Category.STYLE,
    Rule.BACKSLASH_SPACING: Category.STYLE,
    Rule.EXPR_FORMAT: Category.STYLE,
    Rule.SPACES_IN_BRACES: Category.STYLE,
    Rule.REDEFINED_BUILTIN: Category.FUNC,
    Rule.COMMAND_ARGS: Category.FUNC,
    Rule.UNBRACED_EXPR: Category.FUNC,
}


class Violation:
    def __init__(self, id: Rule, message: str, pos: Tuple[int, int]):
        self.id = id
        self.message = message
        self.pos = pos

    def __lt__(self, other):
        return self.pos < other.pos

    def str(self, show_category=False):
        line, col = self.pos
        rule = str(self.id)
        if show_category:
            category = str(_CATEGORY_MAP.get(self.id, ""))
            rule = ":".join([category, rule])

        return f"{line}:{col}: {self.message} [{rule}]"

    @classmethod
    def create(cls, id):
        def func(message: str, pos: Tuple[int, int]):
            return cls(id, message, pos)

        return func



import pathlib

import pytest

from tclint.config import get_config, RunConfig, ConfigError
from tclint.violations import Rule

MY_DIR = pathlib.Path(__file__).parent.resolve()


def test_example_config():
    config_path = MY_DIR / "data" / "tclint.toml"
    config = get_config(config_path)

    global_ = config.get_for_path(pathlib.Path())

    assert global_.exclude == ["ignore_me/", "ignore*.tcl", "/ignore_from_here"]
    assert global_.ignore == [
        Rule("spacing"),
        {"path": pathlib.Path("files_with_bad_indent/"), "rules": [Rule("indent")]},
    ]
    assert global_.extensions == ["tcl"]

    assert global_.style_indent == 2
    assert global_.style_line_length == 100
    assert global_.style_allow_aligned_sets is True
    assert global_.style_indent_namespace_eval is False

    group1 = config.get_for_path(pathlib.Path("other_file_group1/file.tcl"))
    assert group1.style_indent == 3
    assert group1.ignore == [Rule("command-args")]
    assert group1.style_line_length == 100

    group2 = config.get_for_path(pathlib.Path("other_file_group2/foo/file.tcl"))
    assert group2.style_indent == 2
    assert group2.style_allow_aligned_sets is False


def test_invalid_rule():
    with pytest.raises(ConfigError):
        RunConfig.from_dict({"ignore": ["asdf"]})


def test_pyproject():
    config = RunConfig.from_pyproject(MY_DIR / "data")
    global_ = config.get_for_path(pathlib.Path())
    assert global_.style_indent == 2
    assert global_.ignore == [Rule("spacing")]



from tclint.syntax_tree import Node


def pytest_assertrepr_compare(op, left, right):
    if isinstance(left, Node) and isinstance(right, Node) and op == "==":
        return ["Syntax trees not equal", *left.diff(right)]



import pathlib

import pytest

from tclint.parser import (
    Parser,
    Script,
    Command,
    Comment,
    CommandSub,
    ArgExpansion,
    BareWord,
    CompoundBareWord,
    QuotedWord,
    BracedWord,
    VarSub,
    List,
    BracedExpression,
    Expression,
    BinaryOp,
    TernaryOp,
    Function,
    TclSyntaxError,
)

MY_DIR = pathlib.Path(__file__).parent.resolve()


def parse(input):
    parser = Parser(debug=True)
    return parser.parse(input)


def test_null():
    script = ""
    tree = parse(script)
    assert tree == Script()


def test_escaped_brace():
    script = r"puts {Dinosaur dan\} {asdf}}"
    tree = parse(script)
    assert tree == Script(
        Command(BareWord("puts"), BracedWord(r"Dinosaur dan\} {asdf}"))
    )


def test_comments():
    script = r"""# lonely comment
    puts "hello"; # comment after command
    # multiline \
    comment
    # comment over \\
    puts "goodbye"
    # comment with a semicolon; hello"""
    tree = parse(script)
    assert tree == Script(
        Comment(" lonely comment"),
        Command(BareWord("puts"), QuotedWord(BareWord("hello"))),
        Comment(" comment after command"),
        Comment(" multiline \\\n    comment"),
        Comment(" comment over \\\\"),
        Command(BareWord("puts"), QuotedWord(BareWord("goodbye"))),
        Comment(" comment with a semicolon; hello"),
    )


def test_proc_in_proc():
    script = """proc proc_in_proc {} {
      proc asdf {} {
        puts "Hello world"
      }
    }"""
    tree = parse(script)

    assert tree == Script(
        Command(
            BareWord("proc"),
            BareWord("proc_in_proc"),
            BracedWord(""),
            Script(
                Command(
                    BareWord("proc"),
                    BareWord("asdf"),
                    BracedWord(""),
                    Script(
                        Command(BareWord("puts"), QuotedWord(BareWord("Hello world"))),
                    ),
                ),
            ),
        )
    )

    outer_proc = tree.children[0]
    inner_proc = outer_proc.args[2].children[0]
    puts_cmd = inner_proc.args[2].children[0]

    assert outer_proc.line == 1
    assert inner_proc.line == 2
    assert puts_cmd.line == 3


def test_arg_expansion():
    script = "puts {*}{foo bar baz}; {*}  "
    tree = parse(script)
    assert tree == Script(
        Command(BareWord("puts"), ArgExpansion(BracedWord("foo bar baz"))),
        Command(BracedWord("*")),
    )


def test_weird_code_block():
    """Test that } that appears to be in comment actually terminates body of
    proc."""

    script = """proc foo {} {
        # bar }
        puts baz
    }"""

    tree = parse(script)
    assert tree == Script(
        Command(
            BareWord("proc"),
            BareWord("foo"),
            BracedWord(""),
            Script(Comment(" bar ")),
        ),
        Command(BareWord("puts"), BareWord("baz")),
        Command(BareWord("}")),
    )


def test_var_sub():
    # TODO: test backslash sub in var name?
    script = r"""puts "Hello $name"
    puts prefix-$suffix
    puts $hElLo__h0w::areyou:::
    puts $:
    puts ${as"{]$l}"""

    tree = parse(script)
    assert tree == Script(
        Command(BareWord("puts"), QuotedWord(BareWord("Hello "), VarSub("name"))),
        Command(
            BareWord("puts"),
            CompoundBareWord(BareWord("prefix-"), VarSub("suffix")),
        ),
        Command(BareWord("puts"), VarSub("hElLo__h0w::areyou:::")),
        Command(BareWord("puts"), BareWord("$:")),
        Command(BareWord("puts"), VarSub('as"{]$l')),
    )


def test_fancy_var_sub():
    script = "$name([calculate index]-middle-$suffix)"
    tree = parse(script)
    assert tree == Script(
        Command(
            VarSub(
                "name",
                CommandSub(Command(BareWord("calculate"), BareWord("index"))),
                BareWord("-middle-"),
                VarSub("suffix"),
            )
        )
    )


def test_command_sub():
    script = r'"hello [puts {[} [nested \] command]]"'
    tree = parse(script)
    assert tree == Script(
        Command(
            QuotedWord(
                BareWord("hello "),
                CommandSub(
                    Command(
                        BareWord("puts"),
                        BracedWord("["),
                        CommandSub(
                            Command(
                                BareWord("nested"),
                                BareWord("\\]"),
                                BareWord("command"),
                            )
                        ),
                    )
                ),
            )
        )
    )


def test_weird_words():
    script = r"""puts "hello {{}"
    puts h"llo
    puts h}{llo"""
    tree = parse(script)
    assert tree == Script(
        Command(BareWord("puts"), QuotedWord(BareWord(r"hello {{}"))),
        Command(BareWord("puts"), BareWord(r'h"llo')),
        Command(BareWord("puts"), BareWord(r"h}{llo")),
    )


def test_multiline():
    script = r'''puts "Multiline \
    Word"'''
    tree = parse(script)

    # The \\n\s+ doesn't get subbed for a space, but that should be okay. The
    # right thing happens with argument parsing, and we don't have anything else
    # that interprets values of words.

    assert tree == Script(
        Command(BareWord("puts"), QuotedWord(BareWord("Multiline \\\n    Word")))
    )


def test_multiline_braces():
    script = r"""if {1} {
        cmd arg1 \
            arg2 \
            arg3
        }"""
    tree = parse(script)

    assert tree == Script(
        Command(
            BareWord("if"),
            BracedExpression(BareWord("1")),
            Script(
                Command(
                    BareWord("cmd"),
                    BareWord("arg1"),
                    BareWord("arg2"),
                    BareWord("arg3"),
                )
            ),
        )
    )


def test_clean_tcl():
    with open(MY_DIR / "data" / "clean.tcl", "r") as f:
        script = f.read()
    tree = parse(script)

    assert tree == Script(
        Command(
            BareWord("for"),
            Script(Command(BareWord("set"), BareWord("i"), BareWord("1"))),
            BracedExpression(
                BinaryOp(
                    VarSub("i"),
                    BareWord("<"),
                    BareWord("100"),
                ),
            ),
            Script(Command(BareWord("incr"), BareWord("i"))),
            Script(
                Command(
                    BareWord("if"),
                    BracedExpression(
                        BinaryOp(
                            VarSub("i"),
                            BareWord("%"),
                            BinaryOp(
                                BareWord("15"),
                                BareWord("=="),
                                BareWord("0"),
                            ),
                        ),
                    ),
                    Script(
                        Command(BareWord("puts"), QuotedWord(BareWord("FizzBuzz"))),
                    ),
                    BareWord("elseif"),
                    BracedExpression(
                        BinaryOp(
                            VarSub("i"),
                            BareWord("%"),
                            BinaryOp(
                                BareWord("3"),
                                BareWord("=="),
                                BareWord("0"),
                            ),
                        ),
                    ),
                    Script(
                        Command(BareWord("puts"), QuotedWord(BareWord("Fizz"))),
                    ),
                    BareWord("elseif"),
                    BracedExpression(
                        BinaryOp(
                            VarSub("i"),
                            BareWord("%"),
                            BinaryOp(
                                BareWord("5"),
                                BareWord("=="),
                                BareWord("0"),
                            ),
                        ),
                    ),
                    Script(
                        Command(BareWord("puts"), QuotedWord(BareWord("Buzz"))),
                    ),
                    BareWord("else"),
                    Script(
                        Command(BareWord("puts"), VarSub("i")),
                    ),
                ),
            ),
        ),
    )


def test_line_numbers():
    script = r"""# line 1
    # line 2
    if {1} {
        # line 4
    }
    # line 6
    """
    tree = parse(script)
    assert tree.children[0].line == 1
    assert tree.children[0].col == 1
    assert tree.children[1].line == 2
    # weird col numbers on account of extra spaces in multiline string
    assert tree.children[1].col == 5
    if_cmd = tree.children[2]
    if_body = if_cmd.children[2]
    assert if_body.children[0].line == 4
    assert if_body.children[0].col == 9
    assert tree.children[3].line == 6

    # col 5 b/c of extra whitespace in multiline string
    assert tree.end_pos == (7, 5)


def test_syntax_error():
    script = 'puts "hello'
    with pytest.raises(TclSyntaxError):
        parse(script)


def test_switch():
    script = r"""switch -regexp $foo { "a" {
            puts "a"
        }
        "b" {
            puts "b"
        }
    }"""

    tree = parse(script)
    assert tree == Script(
        Command(
            BareWord("switch"),
            BareWord("-regexp"),
            VarSub("foo"),
            List(
                QuotedWord(BareWord("a")),
                Script(
                    Command(BareWord("puts"), QuotedWord(BareWord("a"))),
                ),
                QuotedWord(BareWord("b")),
                Script(
                    Command(BareWord("puts"), QuotedWord(BareWord("b"))),
                ),
            ),
        )
    )

    # test positions - this has been tricky
    assert tree.children[0].children[3].children[0].col == 23
    assert tree.children[0].children[3].children[1].children[0].line == 2


def test_other_switch():
    script = r'switch $switchopt -- $foo "a" "puts a" "b" "puts b"'

    tree = parse(script)
    assert tree == Script(
        Command(
            BareWord("switch"),
            VarSub("switchopt"),
            BareWord("--"),
            VarSub("foo"),
            QuotedWord(BareWord("a")),
            Script(
                Command(BareWord("puts"), BareWord("a")),
            ),
            QuotedWord(BareWord("b")),
            Script(
                Command(BareWord("puts"), BareWord("b")),
            ),
        )
    )


def test_puts_blank():
    script = 'puts ""'

    tree = parse(script)
    assert tree == Script(Command(BareWord("puts"), QuotedWord()))


def test_eval_positions():
    script = r"""eval  command  arg1\
 arg2"""

    tree = parse(script)
    assert tree == Script(
        Command(
            BareWord("eval"),
            Script(Command(BareWord("command"), BareWord("arg1"), BareWord("arg2"))),
        )
    )
    eval_command = tree.children[0]
    script = eval_command.args[0]
    command = script.children[0]
    assert command.children[0].pos == (1, 7)
    assert command.children[1].pos == (1, 16)
    assert command.children[2].pos == (2, 2)


def test_eval_braced_multi_arg():
    script = "eval puts {a b c}"

    tree = parse(script)
    assert tree == Script(
        Command(BareWord("eval"), BareWord("puts"), BracedWord("a b c"))
    )

    # This reflects the actual Tcl semantics, but we currently treat this as
    # un-parseable since we don't have a clean way to handle the positions of these
    # items for style-checking purposes.
    # assert tree == Script(
    #     Command(
    #         BareWord("eval"),
    #         Script(
    #             Command(BareWord("puts"), BareWord("a"), BareWord("b"), BareWord("c"))
    #         ),
    #     )
    # )


def test_eval():
    script = "eval {puts {a b c}}"

    tree = parse(script)
    assert tree == Script(
        Command(
            BareWord("eval"), Script(Command(BareWord("puts"), BracedWord("a b c")))
        )
    )


def test_dict_for():
    script = r"""dict for {key value} mydict {
        puts "$key $value"
    }"""

    tree = parse(script)
    assert tree == Script(
        Command(
            BareWord("dict"),
            BareWord("for"),
            BracedWord("key value"),
            BareWord("mydict"),
            Script(
                Command(
                    BareWord("puts"),
                    QuotedWord(VarSub("key"), BareWord(" "), VarSub("value")),
                ),
            ),
        )
    )


def test_eval_lines():
    script = r"""namespace eval my_namespace {
        puts "asdf"
    }"""

    tree = parse(script)
    puts_cmd = tree.children[0].children[3].children[0]
    assert puts_cmd.line == 2


def test_recursive_parse_in_cmd_sub():
    script = "[catch {analyze_power_grid -net $net -corner $corner} err]"
    tree = parse(script)

    assert tree == Script(
        Command(
            CommandSub(
                Command(
                    BareWord("catch"),
                    Script(
                        Command(
                            BareWord("analyze_power_grid"),
                            BareWord("-net"),
                            VarSub("net"),
                            BareWord("-corner"),
                            VarSub("corner"),
                        )
                    ),
                    BareWord("err"),
                )
            )
        )
    )


def test_parse_list():
    val = "alpha beta gamma"
    node = BracedWord(val, pos=(1, 1), end_pos=(1, 1 + len(val)))
    parser = Parser(debug=True)
    list_node = parser.parse_list(node)
    assert list_node.children == [
        BareWord("alpha"),
        BareWord("beta"),
        BareWord("gamma"),
    ]


def test_expr_simple():
    """A single word without substitution should parse properly as an expression
    even without braces."""
    script = 'expr "5"'
    tree = parse(script)
    assert tree == Script(Command(BareWord("expr"), Expression(BareWord("5"))))


def test_expr_sub_brace():
    script = "expr {int($foo)}"
    tree = parse(script)
    assert tree == Script(
        Command(
            BareWord("expr"),
            BracedExpression(Function(BareWord("int"), VarSub("foo"))),
        )
    )


def test_expr_sub_no_brace():
    """Since this isn't wrapped in {...}, should silently parse as normal Tcl.

    (flagging the fact it's not being parsed as an expr would get handled by a
    separate lint check)
    """
    script = "expr int($foo)"
    tree = parse(script)
    assert tree == Script(
        Command(
            BareWord("expr"),
            CompoundBareWord(BareWord("int("), VarSub("foo"), BareWord(")")),
        )
    )


def test_expr_finite_check():
    """From bottom of https://wiki.tcl-lang.org/page/Inf."""
    script = "expr {[string is double -strict $x] && $x == $x && $x + 1 != $x}"
    tree = parse(script)
    assert tree == Script(
        Command(
            BareWord("expr"),
            BracedExpression(
                BinaryOp(
                    CommandSub(
                        Command(
                            BareWord("string"),
                            BareWord("is"),
                            BareWord("double"),
                            BareWord("-strict"),
                            VarSub("x"),
                        )
                    ),
                    BareWord("&&"),
                    BinaryOp(
                        VarSub("x"),
                        BareWord("=="),
                        BinaryOp(
                            VarSub("x"),
                            BareWord("&&"),
                            BinaryOp(
                                VarSub("x"),
                                BareWord("+"),
                                BinaryOp(BareWord("1"), BareWord("!="), VarSub("x")),
                            ),
                        ),
                    ),
                ),
            ),
        )
    )


def test_expr_newline():
    """Mostly meant to test backslash newline, also sneaks in ternary, different
    word types, and indexed varsub."""
    script = r"""expr {"conditional" ? $::env(FOO) : \
        {foo}}"""
    tree = parse(script)
    assert tree == Script(
        Command(
            BareWord("expr"),
            BracedExpression(
                TernaryOp(
                    QuotedWord(BareWord("conditional")),
                    BareWord("?"),
                    VarSub("::env", BareWord("FOO")),
                    BareWord(":"),
                    BracedWord("foo"),
                ),
            ),
        )
    )


def test_expr_no_spaces_binop():
    script = "expr {1-1}; expr {1eq1};"
    tree = parse(script)
    assert tree == Script(
        Command(
            BareWord("expr"),
            BracedExpression(BinaryOp(BareWord("1"), BareWord("-"), BareWord("1"))),
        ),
        Command(
            BareWord("expr"),
            BracedExpression(BinaryOp(BareWord("1"), BareWord("eq"), BareWord("1"))),
        ),
    )


def test_newline_in_expr():
    script = """expr {$foo eq "foo" &&
        $bar eq "bar"}"""
    tree = parse(script)
    assert tree == Script(
        Command(
            BareWord("expr"),
            BracedExpression(
                BinaryOp(
                    VarSub("foo"),
                    BareWord("eq"),
                    BinaryOp(
                        QuotedWord(BareWord("foo")),
                        BareWord("&&"),
                        BinaryOp(
                            VarSub("bar"),
                            BareWord("eq"),
                            QuotedWord(BareWord("bar")),
                        ),
                    ),
                ),
            ),
        )
    )


def test_subparsed_positions():
    script = r"""if {1} pwd
if 1 {
    pwd
}"""
    tree = parse(script)

    if0 = tree.children[0]
    if0_expr = if0.args[0]
    if0_expr_inner = if0_expr.children[0]
    assert if0_expr_inner.pos == (1, 5)
    if0_body = if0.args[1]
    if0_body_inner = if0_body.children[0]
    assert if0_body_inner.pos == (1, 8)

    if1 = tree.children[1]
    if1_expr = if1.args[0]
    if1_expr_inner = if1_expr.children[0]
    assert if1_expr_inner.pos == (2, 4)
    if1_body = if1.args[1]
    if1_body_inner = if1_body.children[0]
    assert if1_body_inner.pos == (3, 5)



from importlib_metadata import entry_points
import json
import pathlib
from typing import Dict, Optional, List
from types import ModuleType


class _PluginManager:
    def __init__(self):
        self._loaded = {}
        self._installed = {}
        self._command_specs = {}
        for plugin in entry_points(group="tclint.plugins"):
            if plugin.name in self._installed:
                print(f"Warning: found duplicate definitions for plugin {plugin.name}")
            self._installed[plugin.name] = plugin

    def load(self, name: str) -> Optional[Dict]:
        if name in self._loaded:
            return self._loaded[name]

        mod = self._load(name)
        self._loaded[name] = mod
        return mod

    def get_mod(self, name: str) -> Optional[ModuleType]:
        if name not in self._installed:
            print(f"Warning: plugin {name} could not be found")
            return None

        plugin = self._installed[name]

        try:
            module = plugin.load()
        except Exception as e:
            print(f"Warning: error loading plugin {name}: {e}")
            return None

        return module

    def load_command_specs(self, command_specs: List[pathlib.Path]):
        for spec_file in command_specs:
            try:
                with open(spec_file, "r") as f:
                    spec = json.load(f)
            except FileNotFoundError:
                print(f"Spec file {spec_file} not found, skipping...")
                continue

            try:
                plugin_name = spec["plugin"]
            except KeyError:
                print(f"Invalid spec file {spec_file}, missing key 'plugin'")
                continue

            try:
                command_spec = spec["spec"]
            except KeyError:
                print(f"Invalid spec file {spec_file}, missing key 'spec'")
                continue

            if plugin_name in self._command_specs:
                print(
                    f"Warning: overwriting existing spec for {plugin_name} with"
                    f" {spec_file}"
                )
            self._command_specs[plugin_name] = command_spec

    def _load(self, name: str):
        module = self.get_mod(name)
        if module is None:
            print(f"Skipping requested plugin {name}")
            return None

        if name in self._command_specs:
            if not hasattr(module, "commands_from_spec"):
                print(
                    f"Warning: skipping plugin {name}: associated spec file provided"
                    " but plugin doesn't support spec files"
                )
                return None

            commands_from_spec = getattr(module, "commands_from_spec")
            return commands_from_spec(self._command_specs[name])

        if not hasattr(module, "commands"):
            print(
                f"Warning: skipping plugin {name} since no spec was provided and it"
                " does not define commands"
            )
            return None

        return getattr(module, "commands")


# TODO: we'll probably want to construct this in the tclint entry point and pass
# it around rather than using a singleton instance, but this made for an easier
# refactor.
PluginManager = _PluginManager()



import re

from tclint.commands import get_commands
from tclint.violations import Rule, Violation

from tclint.syntax_tree import Visitor, Script, List, ParenExpression, BracedExpression


class IndentLevelChecker(Visitor):
    """Checks that each line is indented at the correct level.

    Reports 'indent' violations.
    """

    def __init__(self):
        # -1 lets us add 1 in all script visitors and get the correct level in
        # the top-level script
        self.level = -1

        # maps lines to expected indent levels
        self.expected_levels = {}

    def check(self, input, tree, config):
        self._is_namespace_eval = False
        self._indent_namespace_eval = config.style_indent_namespace_eval

        # first have visitor calculate expected indent on each line
        tree.accept(self, recurse=False)

        # then iterate line-by-line to verify that actual indentation matches
        violations = []
        for lineno, line in enumerate(input.split("\n")):
            # 1-indexed
            lineno += 1

            if lineno not in self.expected_levels:
                continue
            if self.expected_levels[lineno] is None:
                continue

            expected_level = self.expected_levels[lineno]
            indent = "\t" if config.style_indent == "tab" else " " * config.style_indent
            expected_indent = indent * expected_level

            actual_indent = ""
            for c in line:
                if c in {" ", "\t"}:
                    actual_indent += c
                else:
                    break

            if expected_indent != actual_indent:
                if config.style_indent == "tab":
                    expected_str = (
                        f"{expected_level} tab{'s' if expected_level != 1 else ''}"
                    )
                else:
                    count = expected_level * config.style_indent
                    expected_str = f"{count} space{'s' if count != 1 else ''}"

                if " " in actual_indent and "\t" in actual_indent:
                    actual_str = "mix of tabs and spaces"
                elif "\t" in actual_indent:
                    actual_str = f"{len(actual_indent)}"
                    if config.style_indent != "tab":
                        actual_str += f" tab{'s' if len(actual_indent) != 1 else ''}"
                else:
                    actual_str = f"{len(actual_indent)}"
                    if config.style_indent == "tab":
                        actual_str += f" space{'s' if len(actual_indent) != 1 else ''}"

                pos = (lineno, 1)
                violations.append(
                    Violation(
                        Rule.INDENT,
                        f"expected indent of {expected_str}, got {actual_str}",
                        pos,
                    )
                )

        return violations

    def visit_script(self, script):
        should_indent = not self._is_namespace_eval or self._indent_namespace_eval
        self._visit_block(script, should_indent=should_indent)

    def visit_list(self, list):
        # lists work a bit like scripts. However, they only add a level of
        # indentation when there's a newline between the end and start of an
        # item in the list (or between the start of the list and the first
        # element). This seems to be a good standard for enforcing conventional
        # formatting in a couple different contexts, e.g. switch and apply. See
        # tests/data/indent_lists.tcl for test cases that exercise this logic.

        should_indent = False

        prev_line = list.line
        for child in list.children:
            if prev_line is not None and prev_line != child.line:
                should_indent = True
                break
            prev_line = child.end_pos[0]

        self._visit_block(list, should_indent=should_indent)

    def _visit_block(self, block, should_indent=True):
        if should_indent:
            self._incr_level()
        for child in block.children:
            child.accept(self)
        if should_indent:
            self._decr_level()

        # check indentation of closing token of script
        if block.end_pos:
            end_line, _ = block.end_pos
            if end_line not in self.expected_levels:
                self.expected_levels[end_line] = self.level

    def visit_command(self, command):
        assert len(command.children) > 0

        if command.line in self.expected_levels:
            # already visited something on this line
            # TODO: do we want to return here? do we still wanna check continuation?
            return

        prev_namespace_eval = self._is_namespace_eval
        self._is_namespace_eval = (
            command.routine == "namespace"
            and len(command.args) > 0
            and command.args[0].contents == "eval"
        )

        self.expected_levels[command.line] = self.level

        for child in command.children:
            # continuations of command must be indented. If the node is a list
            # or script however, the indentation only applies to contents of that node
            # (handled by the relevant visitors)
            if not isinstance(child, (Script, List)):
                self._incr_level()

            child.accept(self)

            if not isinstance(child, (Script, List)):
                self._decr_level()

        self._is_namespace_eval = prev_namespace_eval

    def visit_comment(self, comment):
        self._set_level(comment)

    def visit_command_sub(self, command_sub):
        self._set_level(command_sub)

    def visit_bare_word(self, word):
        self._set_level(word)

    def visit_braced_word(self, word):
        self._set_level(word)

    def visit_quoted_word(self, word):
        self._set_level(word)

    def visit_compound_bare_word(self, word):
        self._set_level(word)

    def visit_var_sub(self, var_sub):
        self._set_level(var_sub)

    def visit_arg_expansion(self, arg_expansion):
        self._set_level(arg_expansion)

    def visit_expression(self, expression):
        self._visit_expression(expression)

    def visit_braced_expression(self, expression):
        self._visit_expression(expression)

    def _visit_expression(self, expression):
        prev_level = self.level
        # don't check indents for things under expression
        self.level = None
        for child in expression.children:
            child.accept(self, recurse=True)
        self.level = prev_level

    def _set_level(self, node):
        if node.line in self.expected_levels:
            return

        self.expected_levels[node.line] = self.level

    def _incr_level(self):
        if self.level is None:
            return
        self.level += 1

    def _decr_level(self):
        if self.level is None:
            return
        self.level -= 1


class SpacingChecker(Visitor):
    """Ensures each word of a command is separated by a single space (if on the
    same line).

    Reports 'spacing' violations.
    """

    def __init__(self):
        self.violations = []
        self.set_groups = [
            []
            # [set cmd...]
        ]

    def check(self, _, tree, config):
        self.aligned_set = config.style_allow_aligned_sets

        tree.accept(self, recurse=True)

        if self.aligned_set:
            self._check_set_spacing()

        return self.violations

    def visit_command(self, command):
        if len(command.args) == 0:
            return

        handle_as_aligned_set = command.routine == "set" and self.aligned_set

        last_word = command.children[0]
        for i, word in enumerate(command.children[1:]):
            if last_word.end_pos[0] != word.pos[0]:
                # ignore if on diff lines
                last_word = word
                continue

            if handle_as_aligned_set and i == 1:
                # relies on visit_command being called in order
                if (
                    len(self.set_groups[-1]) > 0
                    and command.line == self.set_groups[-1][-1].line + 1
                ):
                    self.set_groups[-1].append(command)
                else:
                    self.set_groups.append([command])

                continue

            spacing = word.pos[1] - last_word.end_pos[1]
            if spacing != 1:
                self.violations.append(
                    Violation(
                        Rule.SPACING,
                        f"expected 1 space between words, got {spacing}",
                        last_word.end_pos,
                    )
                )
            last_word = word

    def _check_set_spacing(self):
        for group in self.set_groups:
            furthest_value_start = 0
            for cmd in group:
                value_start = cmd.args[1].pos[1]
                if value_start > furthest_value_start:
                    furthest_value_start = value_start

            violations = []
            all_aligned = True
            for cmd in group:
                if cmd.args[1].pos[1] != furthest_value_start:
                    all_aligned = False

                spacing = cmd.args[1].pos[1] - cmd.args[0].end_pos[1]
                if spacing != 1:
                    violations.append(
                        Violation(
                            Rule.SPACING,
                            "expected 1 space between value and name or fully aligned"
                            " block",
                            cmd.args[1].pos,
                        )
                    )

            # If all set values are aligned, waive violations. Alignment is
            # meaningless if there's only one command in the group, always
            # report in that case.
            if len(group) == 1 or not all_aligned:
                self.violations.extend(violations)

    def visit_binary_op(self, binary_op):
        return self._check_op_spacing(binary_op)

    def visit_ternary_op(self, ternary_op):
        return self._check_op_spacing(ternary_op)

    def _check_op_spacing(self, expression):
        last_child = expression.children[0]
        for child in expression.children[1:]:
            if last_child.end_pos[0] != child.pos[0]:
                # ignore if on diff lines
                last_child = child
                continue

            # TODO: collapse to one violation for multiple violations in one expression
            spacing = child.pos[1] - last_child.end_pos[1]
            if spacing != 1:
                self.violations.append(
                    Violation(
                        Rule.EXPR_FORMAT,
                        "expected 1 space between operands and operators in expression",
                        last_child.end_pos,
                    )
                )

            last_child = child

    def visit_paren_expression(self, expression):
        assert len(expression.children) == 1
        child = expression.children[0]

        if isinstance(child, ParenExpression):
            # check for doubly-nested parens, e.g. `((foo))`
            self.violations.append(
                Violation(
                    Rule.EXPR_FORMAT,
                    "unecessary set of parentheses around expression",
                    expression.pos,
                )
            )

        # TODO: collapse to one violation if both errors present
        if (
            expression.pos[0] == child.pos[0]
            and (child.pos[1] - expression.pos[1]) != 1
        ):
            self.violations.append(
                Violation(
                    Rule.EXPR_FORMAT,
                    "expected no space between open paren and contained expression",
                    expression.pos,
                )
            )
        if (
            expression.end_pos[0] == child.end_pos[0]
            and (expression.end_pos[1] - child.end_pos[1]) != 1
        ):
            self.violations.append(
                Violation(
                    Rule.EXPR_FORMAT,
                    "expected no space between close paren and contained expression",
                    child.end_pos,
                )
            )

    def visit_unary_op(self, unary_op):
        assert len(unary_op.children) == 2
        operator = unary_op.children[0]
        operand = unary_op.children[1]

        if operator.line != operand.line:
            self.violations.append(
                Violation(
                    Rule.EXPR_FORMAT,
                    "expected unary operator and operand to be on the same line",
                    unary_op.children[0].end_pos,
                )
            )

        spacing = operand.pos[1] - operator.end_pos[1]
        if spacing != 0:
            self.violations.append(
                Violation(
                    Rule.EXPR_FORMAT,
                    "expected no space between unary operator and operand",
                    unary_op.children[0].end_pos,
                )
            )

    def visit_function(self, function):
        """Function must be formatted like func(arg1, arg2, ...)"""
        last_child = function.children[0]
        for i, child in enumerate(function.children[1:]):
            if child.pos[0] != last_child.end_pos[0]:
                # ignore spacing if not on same line
                last_child = child
                continue

            spacing = child.pos[1] - last_child.end_pos[1]
            if i == 0 and spacing != 1:
                self.violations.append(
                    Violation(
                        Rule.EXPR_FORMAT,
                        "expected no space between opening paren of function and first"
                        " argument",
                        last_child.end_pos,
                    )
                )
            elif i % 2 == 0 and spacing != 1:
                self.violations.append(
                    Violation(
                        Rule.EXPR_FORMAT,
                        "expected 1 space between function arguments",
                        last_child.end_pos,
                    )
                )
            elif i % 2 == 1 and spacing != 0:
                self.violations.append(
                    Violation(
                        Rule.EXPR_FORMAT,
                        "expected no space between argument and following comma",
                        last_child.end_pos,
                    )
                )
            last_child = child

        if (function.end_pos[0] == last_child.end_pos[0]) and (
            function.end_pos[1] - last_child.end_pos[1] != 1
        ):
            self.violations.append(
                Violation(
                    Rule.EXPR_FORMAT,
                    "expected no space between final argument and closing paren of"
                    " function",
                    last_child.end_pos,
                )
            )


class BackslashNewlineChecker(Visitor):
    """Checks that backslash newline substitutions are spaced properly. Reports
    'backslash-spacing' violations."""

    def check(self, input, tree, config):
        self._excluded_ranges = []
        tree.accept(self, recurse=True)

        violations = []
        lines = input.split("\n")
        for min, max in self._excluded_ranges:
            for i in range(min, max):
                lines[i - 1] = None

        for i, line in enumerate(lines):
            if line is None:
                continue

            lineno = i + 1
            # TODO: whitespace after a backslash could be an error. It'll be
            # flagged as a trailing-whitespace violation, but maybe it's worth
            # flagging something more distinct?
            line = line.rstrip(" \t")
            if line.endswith("\\"):
                # count whitespace before \
                whitespace_count = 0
                for c in reversed(line[:-1]):
                    if c not in {" ", "\t"}:
                        break
                    whitespace_count += 1

                if whitespace_count == len(line) - 1:
                    # entire line consists of whitespace + \
                    # TODO: consider flagging violation here? this is probably
                    # not a necessary "\". could also be worth checking
                    # indentation of a lonely "\", if we allow them
                    continue
                if whitespace_count != 1 or line[-2] == "\t":
                    violations.append(
                        Violation(
                            Rule.BACKSLASH_SPACING,
                            "expected 1 space between line contents and backslash",
                            (lineno, len(line)),
                        )
                    )

        return violations

    def visit_braced_word(self, word):
        self._excluded_ranges.append((word.pos[0], word.end_pos[0]))

    def visit_quoted_word(self, word):
        self._excluded_ranges.append((word.pos[0], word.end_pos[0]))


class LineChecker:
    """Ensures lines aren't too long and do not include trailing whitespace.

    Reports 'line-length' and 'trailing-whitespace' violations.
    """

    # ref: https://github.com/eslint/eslint/blob/b29a16b22f234f6134475efb6c7be5ac946556ee/lib/rules/max-len.js#L101 # noqa: E501
    # ^ ironic lint waiver...
    URL_RE = re.compile(r"[^:/?#]:\/\/[^?#]")

    def check(self, input, _, config):
        violations = []
        for i, line in enumerate(input.split("\n")):
            if self.URL_RE.search(line) is not None:
                # ignore URLs
                continue

            lineno = i + 1
            if len(line) > config.style_line_length:
                pos = (lineno, config.style_line_length + 1)
                violations.append(
                    Violation(
                        Rule.LINE_LENGTH,
                        f"line length is {len(line)}, maximum allowed is"
                        f" {config.style_line_length}",
                        pos,
                    )
                )

            if line.endswith((" ", "\t")):
                pos = (lineno, len(line))
                violations.append(
                    Violation(
                        Rule.TRAILING_WHITESPACE, "line has trailing whitespace", pos
                    )
                )

        return violations


class RedefinedBuiltinChecker(Visitor):
    """Ensures names of built-in commands aren't reused by proc definitions.

    Reports 'redefined-builtin' violations.
    """

    def check(self, _, tree, config):
        self._violations = []

        builtin_commands = get_commands(config.command_plugins)
        self._commands = builtin_commands.keys()

        tree.accept(self, recurse=True)

        return self._violations

    def visit_command(self, command):
        if command.routine != "proc":
            return

        name = command.args[0].contents

        if name in self._commands:
            self._violations.append(
                Violation(
                    Rule.REDEFINED_BUILTIN,
                    f"redefinition of built-in command '{name}'",
                    command.pos,
                )
            )


class BlankLineChecker(Visitor):
    """Ensures that there aren't too many blank lines between commands/comments.

    Reports 'blank-lines' violations.
    """

    def check(self, _, tree, config):
        self._violations = []
        self._non_blank_ranges = []
        self._parent_is_block = False
        self._max_blank_lines = config.style_max_blank_lines
        tree.accept(self)
        return self._violations

    def visit_script(self, script):
        self._handle_block(script)

    def visit_comment(self, comment):
        self._handle_item(comment)

    def visit_command(self, command):
        self._handle_item(command)

    def visit_command_sub(self, command_sub):
        self._handle_item(command_sub)

    def visit_quoted_word(self, word):
        self._handle_item(word)

    def visit_compound_bare_word(self, word):
        self._handle_item(word)

    def visit_compound_var_sub(self, var_sub):
        self._handle_item(var_sub)

    def visit_arg_expansion(self, arg_expansion):
        self._handle_item(arg_expansion)

    def visit_list(self, list):
        self._handle_block(list)

    def _handle_block(self, block):
        # every time we visit a script/list, we record the extents of all
        # items within the first level of the block, and then check that
        # the gaps between them don't exceed max_blank_lines.

        # recording the full extent of an item  ensures that we don't flag
        # blank lines within e.g. a quoted multi-line string. we can then
        # recursively perform this check within any nested script arguments to
        # ensure we do check things like if statement bodies.

        # need to handle this case for switch commands, where we have a Script
        # node directly under a List node (which is a block)
        if self._parent_is_block:
            self._non_blank_ranges.append((block.pos[0], block.end_pos[0]))

        # we need to save and restore _non_blank_ranges to facilitate the recursion.
        prev_ranges = self._non_blank_ranges

        self._non_blank_ranges = [
            # (start_line, end_line)
        ]

        for child in block.children:
            self._parent_is_block = True
            child.accept(self)

        last_non_blank_line = block.pos[0]
        for start_line, end_line in self._non_blank_ranges:
            blank_lines = (start_line - last_non_blank_line) - 1

            if blank_lines > self._max_blank_lines:
                self._violations.append(
                    Violation(
                        Rule.BLANK_LINES,
                        f"found {blank_lines} blank lines,"
                        f" expected no more than {self._max_blank_lines}",
                        (last_non_blank_line + 1, 1),
                    )
                )
            last_non_blank_line = end_line

        blank_lines = (block.end_pos[0] - last_non_blank_line) - 1
        if blank_lines > self._max_blank_lines:
            self._violations.append(
                Violation(
                    Rule.BLANK_LINES,
                    f"found {blank_lines} blank lines,"
                    f" expected no more than {self._max_blank_lines}",
                    (last_non_blank_line + 1, 1),
                )
            )

        self._non_blank_ranges = prev_ranges

    def _handle_item(self, item):
        if self._parent_is_block:
            self._non_blank_ranges.append((item.pos[0], item.end_pos[0]))
        for child in item.children:
            self._parent_is_block = False
            child.accept(self)


class SpacesInBracesChecker(Visitor):
    def check(self, _, tree, config):
        self._violations = []
        self._spaces_in_braces = config.style_spaces_in_braces
        tree.accept(self, recurse=True)
        return self._violations

    def visit_script(self, script):
        if not script.braced:
            # We tag whether a script is braced using this member. It's a bit of
            # a hack - it would probably be better to have a BracedScript
            # analagous to BracedExpression, but that change is more invasive.
            #
            # We can distinguish between a braced and bare script using
            # positions, but this extra info is important so we don't check
            # quoted scripts.
            return

        return self._check_braced_arg(script)

    def visit_braced_expression(self, expression):
        return self._check_braced_arg(expression)

    def _check_braced_arg(self, node):
        if len(node.children) == 0:
            # TODO: consider enforcing {} vs { }
            return

        if self._spaces_in_braces:
            expected_spacing = 2
            message = "expected 1 space between contents and enclosing braces"
        else:
            expected_spacing = 1
            message = "expected no space between contents and enclosing braces"

        # check violation at beginning
        first = node.children[0]
        if (
            first.pos[0] == node.pos[0]
            and first.pos[1] - node.pos[1] != expected_spacing
        ):
            self._violations.append(
                Violation(
                    Rule.SPACES_IN_BRACES,
                    message,
                    node.pos,
                )
            )
            return

        # check violation at end
        last = node.children[-1]
        if (
            node.end_pos[0] == last.end_pos[0]
            and node.end_pos[1] - last.end_pos[1] != expected_spacing
        ):
            self._violations.append(
                Violation(
                    Rule.SPACES_IN_BRACES,
                    message,
                    last.end_pos,
                )
            )


class UnbracedExprChecker(Visitor):
    def check(self, _, tree, __):
        self._violations = []
        tree.accept(self, recurse=True)
        return self._violations

    def visit_command(self, command):
        if command.routine != "expr":
            return

        if len(command.args) == 1 and isinstance(command.args[0], BracedExpression):
            return

        for child in command.args:
            if child.contents is None:
                self._violations.append(
                    Violation(
                        Rule.UNBRACED_EXPR,
                        "expression with substitutions should be enclosed by braces",
                        command.args[0].pos,
                    )
                )
                return


def get_checkers():
    return (
        IndentLevelChecker(),
        SpacingChecker(),
        LineChecker(),
        RedefinedBuiltinChecker(),
        BlankLineChecker(),
        BackslashNewlineChecker(),
        SpacesInBracesChecker(),
        UnbracedExprChecker(),
    )
